# Kyler Olsen
# Feb 2024

from enum import Enum
from typing import ClassVar, Sequence
from textwrap import indent


class FileInfo:

    _filename: str
    _line: int
    _col: int
    _length: int
    _lines: int

    def __init__(
        self,
        filename: str,
        line: int,
        col: int,
        length: int,
        lines: int = 0,
    ):
        self._filename = filename
        self._line = line
        self._col = col
        self._length = length
        self._lines = lines

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}"
            f"('{self._filename}',{self._line},{self._col},{self._length})"
        )

    def __str__(self) -> str:
        return f"Ln {self.line}, Col {self.col} in file {self.filename}"

    def __add__(self, other: "FileInfo") -> "FileInfo":
        filename = self.filename
        line = self.line
        col = self.col
        if self.line != other.line:
            if other.lines == 0:
                length = other.col + other.length
            else:
                length = other.length
            lines = other.line - self.line
        else:
            length = (other.col + other.length) - col
            lines = 0
        return FileInfo(
            filename,
            line,
            col,
            length,
            lines,
        )

    @property
    def filename(self) -> str: return self._filename
    @property
    def line(self) -> int: return self._line
    @property
    def col(self) -> int: return self._col
    @property
    def length(self) -> int: return self._length
    @property
    def lines(self) -> int: return self._lines


class CompilerError(Exception):

    _compiler_error_type = "Compiler"

    def __init__(
            self,
            message: str,
            file_info: FileInfo,
            file_info_context: FileInfo | None = None,
        ):
        new_message = message
        new_message += (
            f"\nIn file {file_info.filename} at line {file_info.line} "
        )
        if file_info_context is not None and file_info_context.lines:
            file_info_context = None
        if file_info.lines:
            new_message += f"to line {file_info.line + file_info.lines}"
            with open(file_info.filename, 'r') as file:
                new_message += ''.join(
                    file.readlines()[
                        file_info.line-1:file_info.line + file_info.lines])
        else:
            new_message += f"col {file_info.col}\n\n"
            with open(file_info.filename, 'r') as file:
                new_message += file.readlines()[file_info.line-1]
            if file_info_context is not None:
                context_line = [' '] * max(
                    file_info.col + file_info.length,
                    file_info_context.col +file_info_context.length,
                )
                for i in range(
                    file_info_context.col - 1,
                    file_info_context.col + file_info_context.length
                ):
                    context_line[i] = '~'
                for i in range(
                    file_info.col - 1,
                    file_info.col + file_info.length
                ):
                    context_line[i] = '^'
                new_message += ''.join(context_line)
            else:
                new_message += ' ' * (
                    file_info.col - 1) + '^' * file_info.length

        super().__init__(new_message)

    def compiler_error(self) -> str:
        return (
            f"[{self._compiler_error_type} Error] {type(self).__name__}:\n"
            f"{indent(str(self), '   |', lambda _: True)}"
        )


class CompilerWarning(Warning):

    _compiler_warning_type = "Compiler"

    def __init__(self, message: str, file_info: FileInfo):
        new_message = message
        new_message += (
            f"\nIn file {file_info.filename} at line {file_info.line} "
        )
        if file_info.lines:
            new_message += f"to line {file_info.line + file_info.lines}"
            with open(file_info.filename, 'r') as file:
                new_message += ''.join(
                    file.readlines()[
                        file_info.line-1:file_info.line + file_info.lines])
        else:
            new_message += f"col {file_info.col}\n\n"
            with open(file_info.filename, 'r') as file:
                new_message += file.readlines()[file_info.line-1]
            new_message += ' ' * (
                file_info.col - 1) + '^' * file_info.length

        super().__init__(new_message)

    def compiler_error(self) -> str:
        return (
            f"[{self._compiler_warning_type} Warning] {type(self).__name__}:\n"
            f"{indent(str(self), '   |', lambda _: True)}"
        )

# -- Lexical --

class LexerError(CompilerError):

    _compiler_error_type = "Lexical"


class _InterTokenType(Enum):
    Generic = 'Generic'
    Comment = 'Comment'
    Word = 'Word'
    NumberLiteral = 'NumberLiteral'
    Punctuation = 'Punctuation'


class _NumberLiteralType(Enum):
    Number = 'Number'
    Real = 'Real'
    Exp = 'Exp'


_OnlyNewLineTerminatedTokens = (
    _InterTokenType.Comment,
)

_NewLineTerminatedTokens = _OnlyNewLineTerminatedTokens + (
    _InterTokenType.Word,
    _InterTokenType.NumberLiteral,
    _InterTokenType.Punctuation,
)

_ID_Start = "ABCDEFGHIJKLMNOPQRSTUVWXYZ" "abcdefghijklmnopqrstuvwxyz"

_ID_Continue = _ID_Start# + "0123456789"

_Keywords = (
    'screen', 'graph', 'anim', 'img',
)

_Num_Start = "0123456789"

_Num_Start_Next = {
    _NumberLiteralType.Number: {
        '.': _NumberLiteralType.Real,
    },
}

_Num_Continue = {
    _NumberLiteralType.Number: _Num_Start + ".eE_",
    _NumberLiteralType.Real: _Num_Start + "eE_",
    _NumberLiteralType.Exp: _Num_Start + "_",
}

_Num_Continue_Next = {
    _NumberLiteralType.Number: {
        '.': _NumberLiteralType.Real,
        'e': _NumberLiteralType.Exp,
        'E': _NumberLiteralType.Exp,
    },
    _NumberLiteralType.Real: {
        'e': _NumberLiteralType.Exp,
        'E': _NumberLiteralType.Exp,
    },
}

_Punctuation_Any = "_+-*/%^<>=!{[(}]),;:∑∏∞≠≤≥∫αβπ→"

_Punctuation = (
    "+",  "-",  "*",  "/",  "%",  "^",
    "=",  "!",  "<",  "<=", ">",  ">=",
    "{",  "}",  "[",  "]",  "(",  ")",
    "_",  "->", ",",  ";",  ":",  "∑",
    "∏",  "∞",  "≠",  "≤",  "≥",  "∫",
    "α",  "β",  "θ",  "π",  "→",
)

_Punctuation_Conversion = {
    "<=": "≤",
    ">=": "≥",
    "->": "→",
}

_ID_Conversion = {
    "sum": "∑",
    "pi": "π",
    "alpha": "α",
    "beta": "β",
    "inf": "∞",
    "product": "∏",
    "integral": "∫",
    "theta": "θ",
}


class Token:

    _type: ClassVar[str] = 'Generic'
    _value: str
    _file_info: FileInfo

    def __init__(self, value: str, file_info: FileInfo):
        self._value = value
        self._file_info = file_info

    def __str__(self) -> str:
        return f"Type: {self._type}, Value: {self.value}"

    @property
    def value(self) -> str: return self._value

    @property
    def file_info(self) -> FileInfo: return self._file_info

class Directive(Token): _type = 'Directive'
class Identifier(Token): _type = 'Identifier'
class Keyword(Token): _type = 'Keyword'
class NumberLiteral(Token): _type = 'NumberLiteral'
class Punctuation(Token): _type = 'Punctuation'


def lexer(file: str, filename: str) -> Sequence[Token]:
    tokens: list[Token] = []
    current: str = ""
    current_line: int = 0
    current_col: int = 0
    number_type: _NumberLiteralType = _NumberLiteralType.Number
    token_type: _InterTokenType = _InterTokenType.Generic

    for line, line_str in enumerate(file.splitlines()):
        fi = FileInfo(filename, current_line, current_col, len(current))
        if token_type in _NewLineTerminatedTokens:
            if token_type is _InterTokenType.Word:
                if len(current) > 15:
                    raise LexerError("Identifier Too Long", fi)
                if current in _Keywords:
                    tokens.append(Keyword(current, fi))
                elif current.lower() in _ID_Conversion.keys():
                    tokens.append(
                        Punctuation(_ID_Conversion[current.lower()], fi))
                else:
                    tokens.append(Identifier(current, fi))
            elif token_type is _InterTokenType.NumberLiteral:
                tokens.append(NumberLiteral(current, fi))
                number_type = _NumberLiteralType.Number
            elif token_type is _InterTokenType.Punctuation:
                if current not in _Punctuation:
                    raise LexerError("Invalid Punctuation", fi)
                if current in _Punctuation_Conversion.keys():
                    current = _Punctuation_Conversion[current]
                tokens.append(Punctuation(current, fi))
            token_type = _InterTokenType.Generic

        for col, char in enumerate(line_str):
            if token_type in _OnlyNewLineTerminatedTokens:
                current += char
            elif token_type is _InterTokenType.Word:
                if char in _ID_Continue:
                    current += char
                else:
                    fi = FileInfo(
                        filename, current_line, current_col, len(current))
                    if len(current) > 15:
                        raise LexerError("Identifier Too Long", fi)
                    if current in _Keywords:
                        tokens.append(Keyword(current, fi))
                    elif current.lower() in _ID_Conversion.keys():
                        tokens.append(
                            Punctuation(_ID_Conversion[current.lower()], fi))
                    else:
                        tokens.append(Identifier(current, fi))
                    token_type = _InterTokenType.Generic
            elif token_type is _InterTokenType.NumberLiteral:
                if (
                        number_type in _Num_Continue and
                        char in _Num_Continue[number_type]
                    ):
                    current += char
                    if (
                        number_type in _Num_Continue_Next and
                        char in _Num_Continue_Next[number_type]
                    ):
                        number_type = _Num_Continue_Next[number_type][char]
                else:
                    fi = FileInfo(
                        filename, current_line, current_col, len(current))
                    tokens.append(NumberLiteral(current, fi))
                    number_type = _NumberLiteralType.Number
                    token_type = _InterTokenType.Generic
            elif token_type is _InterTokenType.Punctuation:
                if char in _Punctuation_Any and current + char in _Punctuation:
                    current += char
                else:
                    fi = FileInfo(
                        filename, current_line, current_col, len(current))
                    if current not in _Punctuation:
                        raise LexerError("Invalid Punctuation", fi)
                    if current in _Punctuation_Conversion.keys():
                        current = _Punctuation_Conversion[current]
                    tokens.append(Punctuation(current, fi))
                    token_type = _InterTokenType.Generic

            if token_type is _InterTokenType.Generic:
                current = char
                current_line = line + 1
                current_col = col + 1
                if char == '#':
                    token_type = _InterTokenType.Comment
                elif char in _ID_Start:
                    token_type = _InterTokenType.Word
                elif (
                    char == '.' and
                    line_str[col+1] in _Num_Continue[_NumberLiteralType.Real]
                ):
                    token_type = _InterTokenType.NumberLiteral
                    if char in _Num_Start_Next[number_type]:
                        number_type = _Num_Start_Next[number_type][char]
                elif char in _Num_Start:
                    token_type = _InterTokenType.NumberLiteral
                    if char in _Num_Start_Next[number_type]:
                        number_type = _Num_Start_Next[number_type][char]
                elif char in _Punctuation_Any:
                    token_type = _InterTokenType.Punctuation

    fi = FileInfo(filename, current_line, current_col, len(current))
    if token_type in _NewLineTerminatedTokens:
        if token_type is _InterTokenType.Word:
            if len(current) > 31:
                raise LexerError("Identifier Too Long", fi)
            if current in _Keywords:
                tokens.append(Keyword(current, fi))
            elif current.lower() in _ID_Conversion.keys():
                tokens.append(Punctuation(_ID_Conversion[current.lower()], fi))
            else:
                tokens.append(Identifier(current, fi))
        elif token_type is _InterTokenType.NumberLiteral:
            tokens.append(NumberLiteral(current, fi))
            number_type = _NumberLiteralType.Number
        elif token_type is _InterTokenType.Punctuation:
            if current not in _Punctuation:
                raise LexerError("Invalid Punctuation", fi)
            if current in _Punctuation_Conversion.keys():
                current = _Punctuation_Conversion[current]
            tokens.append(Punctuation(current, fi))
        token_type = _InterTokenType.Generic

    return tokens


# # -- Syntax --

# type NestableCodeBlock = ForBlock | WhileBlock | DoBlock | IfBlock

# type Literal = (
#     BuiltInConst |
#     NumberLiteral |
#     CharLiteral |
#     StringLiteral
# )

# type Expression = (
#     Literal |
#     Identifier |
#     UnaryExpression |
#     BinaryExpression |
#     TernaryExpression |
#     FunctionCall |
#     NoOperation
# )

# type Statement = Expression | LetStatement | LoopStatements | NestableCodeBlock

# type DataType = BuiltInDataType | Identifier

# type Operator = UnaryOperator | BinaryOperator | TernaryOperator


# class SyntaxError(CompilerError):

#     _compiler_error_type = "Syntax"


# class UnexpectedEndOfTokenStream(SyntaxError): pass


# class _ExpectedTokenBase(SyntaxError):

#     _token_type = Token

#     def __init__(
#         self,
#         token: Token,
#         expected: str | None = None,
#         found: str | None = None,
#     ):
#         if expected is None:
#             expected = self._token_type.__name__
#             found = found or type(token).__name__
#         else:
#             found = found or token.value
#         message = f"Expected '{expected}' but found '{found}'."
#         super().__init__(message, token.file_info)


# class ExpectedDirective(_ExpectedTokenBase):
#     _type_name = Directive
# class ExpectedIdentifier(_ExpectedTokenBase):
#     _type_name = Identifier
# class ExpectedKeyword(_ExpectedTokenBase):
#     _type_name = Keyword
# class ExpectedNumberLiteral(_ExpectedTokenBase):
#     _type_name = NumberLiteral
# class ExpectedPunctuation(_ExpectedTokenBase):
#     _type_name = Punctuation


# class _UnexpectedTokenBase(_ExpectedTokenBase):

#     def __init__(
#         self,
#         token: Token,
#         expected: str | list[str] | None = None,
#         found: str | None = None,
#     ):
#         if isinstance(expected, list):
#             if len(expected) > 1:
#                 s = ""
#                 for i in expected[:-1]:
#                     s += i + "', '"
#                 s = s[:-1] + "or '" + expected[-1]
#                 expected = s
#             else:
#                 expected = expected[0]
#         super().__init__(token, expected, found)


# class UnexpectedToken(_UnexpectedTokenBase):

#     def __init__(
#         self,
#         token: Token,
#         expected: str | list[str],
#         found: str | None = None,
#     ):
#         if isinstance(expected, list):
#             if len(expected) > 1:
#                 s = ""
#                 for i in expected[:-1]:
#                     s += i + "', '"
#                 s = s[:-1] + "or '" + expected[-1]
#                 expected = s
#         found = found or type(token).__name__
#         super().__init__(token, expected, found)


# class UnexpectedDirective(_UnexpectedTokenBase):
#     _type_name = Directive
# class UnexpectedIdentifier(_UnexpectedTokenBase):
#     _type_name = Identifier
# class UnexpectedKeyword(_UnexpectedTokenBase):
#     _type_name = Keyword
# class UnexpectedNumberLiteral(_UnexpectedTokenBase):
#     _type_name = NumberLiteral
# class UnexpectedPunctuation(_UnexpectedTokenBase):
#     _type_name = Punctuation


# class ExpressionError(Exception): pass


# class LoopStatementsEnum(Enum):
#     ContinueStatement = "continue"
#     BreakStatement = "break"


# class LoopStatements:

#     _content: LoopStatementsEnum
#     _file_info: FileInfo

#     def __init__(
#         self,
#         content: LoopStatementsEnum,
#         file_info: FileInfo,
#     ):
#         self._content = content
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     def __str__(self) -> str: return self._content.value

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} {self._content.value.lower()}\n"
#         return s


# class PostfixUnaryOperatorEnum(Enum):
#     Increment = "++"
#     Decrement = "--"


# class PrefixUnaryOperatorEnum(Enum):
#     Increment = "++"
#     Decrement = "--"
#     Negate = "-"
#     BitwiseNOT = "~"
#     BooleanNOT = "!"
#     AddressOf = "@"
#     Dereference = "$"


# class UnaryOperator:

#     _content: PostfixUnaryOperatorEnum | PrefixUnaryOperatorEnum
#     _file_info: FileInfo

#     def __init__(
#         self,
#         content: PostfixUnaryOperatorEnum | PrefixUnaryOperatorEnum,
#         file_info: FileInfo,
#     ):
#         self._content = content
#         self._file_info = file_info

#     @property
#     def content(self) -> PostfixUnaryOperatorEnum | PrefixUnaryOperatorEnum:
#         return self._content

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     def __str__(self) -> str: return (
#         f"{type(self._content).__name__[:-4]}.{self._content.name}")

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Unary Operator: {self}\n"
#         return s


# class BinaryOperatorEnum(Enum):
#     Addition = "+"
#     Subtraction = "-"
#     Multiplication = "*"
#     Division = "/"
#     Modulus = "%"
#     BitwiseAND = "&"
#     BitwiseOR = "|"
#     BitwiseXOR = "^"
#     LeftShift = "<<"
#     RightShift = ">>"
#     MemberOf = "."
#     Assignment = "="
#     AdditionAssignment = "+="
#     SubtractionAssignment = "-="
#     MultiplicationAssignment = "*="
#     DivisionAssignment = "/="
#     ModulusAssignment = "%="
#     BitwiseANDAssignment = "&="
#     BitwiseORAssignment = "|="
#     BitwiseXORAssignment = "^="
#     LeftShiftAssignment = "<<="
#     RightShiftAssignment = ">>="
#     BooleanAND = "&&"
#     BooleanOR = "||"
#     BooleanXOR = "^^"
#     EqualityComparison = "=="
#     InequalityComparison = "!="
#     LessThan = "<"
#     LessOrEqualToThan = "<="
#     GreaterThan = ">"
#     GreaterOrEqualToThan = ">="


# class BinaryOperator:

#     _content: BinaryOperatorEnum
#     _file_info: FileInfo

#     def __init__(
#         self,
#         content: BinaryOperatorEnum,
#         file_info: FileInfo,
#     ):
#         self._content = content
#         self._file_info = file_info

#     @property
#     def content(self) -> BinaryOperatorEnum: return self._content

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     def __str__(self) -> str: return (
#         f"{type(self._content).__name__[:-4]}.{self._content.name}")

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Binary Operator: {self}\n"
#         return s


# class TernaryOperatorEnum(Enum):
#     TernaryConditional = "?"


# class TernaryOperator:

#     _content: TernaryOperatorEnum
#     _file_info: FileInfo

#     def __init__(
#         self,
#         content: TernaryOperatorEnum,
#         file_info: FileInfo,
#     ):
#         self._content = content
#         self._file_info = file_info

#     @property
#     def content(self) -> TernaryOperatorEnum: return self._content

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     def __str__(self) -> str: return (
#         f"{type(self._content).__name__[:-4]}.{self._content.name}")

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Ternary Operator: {self}\n"
#         return s


# class BuiltInDataTypeEnum(Enum):
#     unsigned = "unsigned"
#     int = "int"
#     fixed = "fixed"
#     float = "float"


# class BuiltInDataType:

#     _content: BuiltInDataTypeEnum
#     _file_info: FileInfo

#     def __init__(
#         self,
#         content: BuiltInDataTypeEnum,
#         file_info: FileInfo,
#     ):
#         self._content = content
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     def __str__(self) -> str: return self._content.value

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Built-In Data Type: {self._content.value}\n"
#         return s


# _Operator_Precedence: tuple[
#     PostfixUnaryOperatorEnum |
#     PrefixUnaryOperatorEnum |
#     BinaryOperatorEnum |
#     TernaryOperatorEnum,
#     ...
# ] = (
#     PrefixUnaryOperatorEnum.AddressOf,
#     PrefixUnaryOperatorEnum.Dereference,
#     PrefixUnaryOperatorEnum.BitwiseNOT,
#     PostfixUnaryOperatorEnum.Decrement,
#     PostfixUnaryOperatorEnum.Increment,
#     PrefixUnaryOperatorEnum.Decrement,
#     PrefixUnaryOperatorEnum.Increment,
#     PrefixUnaryOperatorEnum.Negate,
#     PrefixUnaryOperatorEnum.BooleanNOT,
#     BinaryOperatorEnum.MemberOf,
#     BinaryOperatorEnum.RightShift,
#     BinaryOperatorEnum.LeftShift,
#     BinaryOperatorEnum.BitwiseXOR,
#     BinaryOperatorEnum.BitwiseOR,
#     BinaryOperatorEnum.BitwiseAND,
#     BinaryOperatorEnum.Modulus,
#     BinaryOperatorEnum.Division,
#     BinaryOperatorEnum.Multiplication,
#     BinaryOperatorEnum.Subtraction,
#     BinaryOperatorEnum.Addition,
#     BinaryOperatorEnum.GreaterOrEqualToThan,
#     BinaryOperatorEnum.GreaterThan,
#     BinaryOperatorEnum.LessOrEqualToThan,
#     BinaryOperatorEnum.LessThan,
#     BinaryOperatorEnum.InequalityComparison,
#     BinaryOperatorEnum.EqualityComparison,
#     BinaryOperatorEnum.BooleanXOR,
#     BinaryOperatorEnum.BooleanOR,
#     BinaryOperatorEnum.BooleanAND,
#     TernaryOperatorEnum.TernaryConditional,
#     BinaryOperatorEnum.RightShiftAssignment,
#     BinaryOperatorEnum.LeftShiftAssignment,
#     BinaryOperatorEnum.BitwiseXORAssignment,
#     BinaryOperatorEnum.BitwiseORAssignment,
#     BinaryOperatorEnum.BitwiseANDAssignment,
#     BinaryOperatorEnum.ModulusAssignment,
#     BinaryOperatorEnum.DivisionAssignment,
#     BinaryOperatorEnum.MultiplicationAssignment,
#     BinaryOperatorEnum.SubtractionAssignment,
#     BinaryOperatorEnum.AdditionAssignment,
#     BinaryOperatorEnum.Assignment,
# )


# class NoOperation:

#     _file_info: FileInfo

#     def __init__(
#         self,
#         file_info: FileInfo,
#     ):
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Nop\n"
#         return s


# class Identifier:

#     _content: str
#     _file_info: FileInfo

#     def __init__(
#         self,
#         content: str,
#         file_info: FileInfo,
#     ):
#         self._content = content
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     @property
#     def content(self) -> str: return self._content

#     def __str__(self) -> str: return self._content

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Identifier: {self._content}\n"
#         return s


# class StringLiteral:

#     _content: str
#     _file_info: FileInfo

#     def __init__(
#         self,
#         content: str,
#         file_info: FileInfo,
#     ):
#         self._content = content
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     @property
#     def content(self) -> str: return self._content

#     def __str__(self) -> str: return self._content

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} String Literal: {self._content}\n"
#         return s


# class CharLiteral:

#     _content: str
#     _file_info: FileInfo

#     def __init__(
#         self,
#         content: str,
#         file_info: FileInfo,
#     ):
#         self._content = content
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     @property
#     def content(self) -> str: return self._content

#     def __str__(self) -> str: return self._content

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Character Literal: {self._content}\n"
#         return s


# class NumberLiteral:

#     _content: str
#     _file_info: FileInfo

#     def __init__(
#         self,
#         content: str,
#         file_info: FileInfo,
#     ):
#         self._content = content
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     @property
#     def content(self) -> str: return self._content

#     def __str__(self) -> str: return self._content

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Number Literal: {self._content}\n"
#         return s


# class FunctionArgument:

#     _identifier: Identifier | None
#     _value: Expression
#     _file_info: FileInfo

#     def __init__(
#         self,
#         identifier: Identifier | None,
#         value: Expression,
#         file_info: FileInfo,
#     ):
#         self._identifier = identifier
#         self._value = value
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     @property
#     def identifier(self) -> Identifier | None: return self._identifier

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Function Argument\n"
#         if self._identifier: s += f"{pre_cont}├─ Name: {self._identifier}\n"
#         s += f"{pre_cont}└─ Value: {self._value}\n"
#         return s


# class FunctionCall:

#     _identifier: Identifier
#     _args: list[FunctionArgument]
#     _file_info: FileInfo

#     def __init__(
#         self,
#         identifier: Identifier,
#         args: list[FunctionArgument],
#         file_info: FileInfo,
#     ):
#         self._identifier = identifier
#         self._args = args
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     @property
#     def identifier(self) -> Identifier: return self._identifier

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Function Call: {self._identifier}\n"
#         if self._args:
#             for arg in self._args[:-1]:
#                 s += arg.tree_str(pre_cont + "├─", pre_cont + "│ ")
#             s += self._args[-1].tree_str(pre_cont + "└─", pre_cont + "  ")
#         return s


# class TernaryExpression:

#     _operator: TernaryOperator
#     _operand1: Expression
#     _operand2: Expression
#     _operand3: Expression
#     _file_info: FileInfo

#     def __init__(
#         self,
#         operator: TernaryOperator,
#         operand1: Expression,
#         operand2: Expression,
#         operand3: Expression,
#         file_info: FileInfo,
#     ):
#         self._operator = operator
#         self._operand1 = operand1
#         self._operand2 = operand2
#         self._operand3 = operand3
#         self._file_info = file_info

#     @property
#     def operator(self) -> TernaryOperator: return self._operator

#     @property
#     def operand1(self) -> Expression: return self._operand1

#     @property
#     def operand2(self) -> Expression: return self._operand2

#     @property
#     def operand3(self) -> Expression: return self._operand3

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Ternary Expression: {self._operator}\n"
#         s += self._operand1.tree_str(pre_cont + "├─", pre_cont + "│ ")
#         s += self._operand2.tree_str(pre_cont + "├─", pre_cont + "│ ")
#         s += self._operand3.tree_str(pre_cont + "└─", pre_cont + "  ")
#         return s


# class BinaryExpression:

#     _operator: BinaryOperator
#     _operand1: Expression
#     _operand2: Expression
#     _file_info: FileInfo

#     def __init__(
#         self,
#         operator: BinaryOperator,
#         operand1: Expression,
#         operand2: Expression,
#         file_info: FileInfo,
#     ):
#         self._operator = operator
#         self._operand1 = operand1
#         self._operand2 = operand2
#         self._file_info = file_info

#     @property
#     def operator(self) -> BinaryOperator: return self._operator

#     @property
#     def operand1(self) -> Expression: return self._operand1

#     @property
#     def operand2(self) -> Expression: return self._operand2

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Binary Expression: {self._operator}\n"
#         s += self._operand1.tree_str(pre_cont + "├─", pre_cont + "│ ")
#         s += self._operand2.tree_str(pre_cont + "└─", pre_cont + "  ")
#         return s


# class UnaryExpression:

#     _operator: UnaryOperator
#     _operand: Expression
#     _file_info: FileInfo

#     def __init__(
#         self,
#         operator: UnaryOperator,
#         operand: Expression,
#         file_info: FileInfo,
#     ):
#         self._operator = operator
#         self._operand = operand
#         self._file_info = file_info

#     @property
#     def operator(self) -> UnaryOperator: return self._operator

#     @property
#     def operand(self) -> Expression: return self._operand

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Unary Expression: {self._operator}\n"
#         s += self._operand.tree_str(pre_cont + "└─", pre_cont + "  ")
#         return s


# class LetStatement:

#     _identifier: Identifier
#     _type: DataType
#     _pointer: bool
#     _static: bool
#     _assignment: Literal | None
#     _file_info: FileInfo

#     def __init__(
#         self,
#         identifier: Identifier,
#         type: DataType,
#         pointer: bool,
#         static: bool,
#         assignment: Literal | None,
#         file_info: FileInfo,
#     ):
#         self._identifier = identifier
#         self._type = type
#         self._pointer = pointer
#         self._static = static
#         self._assignment = assignment
#         self._file_info = file_info

#     @property
#     def identifier(self) -> Identifier: return self._identifier

#     @property
#     def assignment(self) -> Literal | None: return self._assignment

#     @property
#     def static(self) -> bool: return self._static

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Let Statement: {self._identifier}\n"
#         s += pre_cont
#         s += '├─ Type: ' if self._assignment else '└─ Type: '
#         if self._static: s+= "static "
#         if self._pointer: s+= "@"
#         s += f"{self._type}\n"
#         if self._assignment is not None:
#             s += f"{pre_cont}└─ Default Value: {self._assignment}\n"
#         return s

#     @staticmethod
#     def _sa(tokens: list[Token], token: Token) -> "LetStatement":
#         start_fi: FileInfo = token.file_info
#         static = token.value == 'static'
#         if static:
#             token = tokens.pop(0)
#             _assert_token(ExpectedKeyword, token, 'let')
#         identifier = tokens.pop(0)
#         _assert_token(ExpectedIdentifier, identifier)
#         token = tokens.pop(0)
#         _assert_token(ExpectedPunctuation, token, ':')
#         pointer, data_type = _data_type_sa(tokens)
#         token = tokens.pop(0)
#         _assert_token(ExpectedPunctuation, token)
#         if token.value not in ['=', ';']:
#             raise UnexpectedPunctuation(token, ['=', ';'])
#         elif token.value == '=':
#             token = tokens.pop(0)
#             _assert_token_literal(token)
#             literal = _literal_map(token) # type: ignore
#             token = tokens.pop(0)
#             _assert_token(ExpectedPunctuation, token)
#             if token.value != ';':
#                 raise UnexpectedPunctuation(token, ';')
#         else: literal = None
#         return LetStatement(
#             Identifier(identifier.value, identifier.file_info),
#             data_type,
#             pointer,
#             static,
#             literal,
#             start_fi + token.file_info,
#         )


# class ElseBlock:

#     _code: list[Statement]
#     _file_info: FileInfo

#     def __init__(
#         self,
#         code: list[Statement],
#         file_info: FileInfo,
#     ):
#         self._code = code[:]
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Else Block\n"
#         if self._code:
#             s += f"{pre_cont}└─ Code\n"
#             for code in self._code[:-1]:
#                 s += code.tree_str(pre_cont + "  ├─", pre_cont + "  │ ")
#             s += self._code[-1].tree_str(pre_cont + "  └─", pre_cont + "    ")
#         return s

#     @staticmethod
#     def _sa(tokens: list[Token]) -> "ElseBlock | None":
#         if tokens and tokens[0].value == 'else':
#             else_token = tokens.pop(0)
#             if tokens[0].value == '{':
#                 else_tokens = _get_nested_group(tokens, ('{','}'))[1]
#                 fi = else_token.file_info + else_tokens[-1].file_info
#                 return ElseBlock(_code_block_sa(else_tokens), fi)
#             else:
#                 statement = _statement_sa(tokens)
#                 fi = else_token.file_info + statement.file_info
#                 return ElseBlock([statement], fi)
#         else:
#             return None


# class ForPreDef:

#     _identifier: Identifier
#     _type: DataType
#     _pointer: bool
#     _assignment: Expression | None
#     _file_info: FileInfo

#     def __init__(
#         self,
#         identifier: Identifier,
#         type: DataType,
#         pointer: bool,
#         assignment: Expression | None,
#         file_info: FileInfo,
#     ):
#         self._identifier = identifier
#         self._type = type
#         self._pointer = pointer
#         self._assignment = assignment
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     @property
#     def identifier(self) -> Identifier: return self._identifier

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} For Loop Pre-Definition: {self._identifier}\n"
#         if self._assignment: s += f"{pre_cont}├─ Type: "
#         else: s += f"{pre_cont}└─ Type: "
#         if self._pointer: s+= "@"
#         s += f"{self._type}\n"
#         if self._assignment: s += f"{pre_cont}└─ Value: {self._assignment}\n"
#         return s


# class ForBlock:

#     _pre_statement: Expression | ForPreDef
#     _condition: Expression
#     _code: list[Statement]
#     _post_statement: Expression
#     _else: ElseBlock | None
#     _file_info: FileInfo

#     def __init__(
#         self,
#         pre_statement: Expression | ForPreDef,
#         condition: Expression,
#         code: list[Statement],
#         post_statement: Expression,
#         else_block: ElseBlock | None,
#         file_info: FileInfo,
#     ):
#         self._pre_statement = pre_statement
#         self._condition = condition
#         self._code = code[:]
#         self._post_statement = post_statement
#         self._else = else_block
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} For Loop\n"
#         if self._code or self._else is not None:
#             cond_pre = f"{pre_cont}├─"
#             cond_pre_cont = f"{pre_cont}│ "
#         else:
#             cond_pre = f"{pre_cont}└─"
#             cond_pre_cont = f"{pre_cont}  "
#         s += f"{cond_pre} Pre-Statement\n"
#         s += self._pre_statement.tree_str(
#             cond_pre_cont + "└─", cond_pre_cont + "  ")
#         s += f"{cond_pre} Condition\n"
#         s += self._condition.tree_str(
#             cond_pre_cont + "└─", cond_pre_cont + "  ")
#         s += f"{cond_pre} Post-Statement\n"
#         s += self._post_statement.tree_str(
#             cond_pre_cont + "└─", cond_pre_cont + "  ")
#         if self._code:
#             if self._else is not None:
#                 s += f"{pre_cont}├─ Code\n"
#                 code_pre = f"{pre_cont}│ "
#             else:
#                 s += f"{pre_cont}└─ Code\n"
#                 code_pre = f"{pre_cont}  "
#             for code in self._code[:-1]:
#                 s += code.tree_str(code_pre + "├─", code_pre + "│ ")
#             s += self._code[-1].tree_str(code_pre + "└─", code_pre + "  ")
#         if self._else is not None:
#             s += self._else.tree_str(pre_cont + "└─", pre_cont + "  ")
#         return s

#     @staticmethod
#     def _sa(tokens: list[Token], stoken: Token) -> "ForBlock":
#         _, three_expressions, closing_parentheses = _get_nested_group(tokens)
#         pre_loop_tokens, semicolon = _get_to_symbol(three_expressions)
#         if (
#             isinstance(pre_loop_tokens[0], Identifier) and
#             pre_loop_tokens[1].value == ':'
#         ):
#             id_token = pre_loop_tokens.pop(0)
#             identifier = Identifier(id_token.value, id_token.file_info)
#             token = pre_loop_tokens.pop(0)
#             _assert_token(ExpectedPunctuation, token, ':')
#             pointer, data_type = _data_type_sa(pre_loop_tokens)
#             if pre_loop_tokens:
#                 token = pre_loop_tokens.pop(0)
#                 _assert_token(ExpectedPunctuation, token, '=')
#                 if not pre_loop_tokens:
#                     fi = semicolon.file_info
#                     raise UnexpectedEndOfTokenStream("Expected Expression.", fi)
#                 pre_loop_expr = _expression_sa(pre_loop_tokens)
#             else:
#                 pre_loop_expr = None
#             if pre_loop_expr is not None:
#                 fi = id_token.file_info + pre_loop_expr.file_info
#             else:
#                 fi = id_token.file_info + data_type.file_info
#             pre_loop = ForPreDef(
#                 identifier,
#                 data_type,
#                 pointer,
#                 pre_loop_expr,
#                 fi,
#             )
#         else:
#             if not pre_loop_tokens:
#                 fi = semicolon.file_info
#                 raise UnexpectedEndOfTokenStream("Expected Expression.", fi)
#             pre_loop = _expression_sa(pre_loop_tokens)
#         loop_condition_tokens, semicolon = _get_to_symbol(three_expressions)
#         if not loop_condition_tokens:
#             fi = semicolon.file_info
#             raise UnexpectedEndOfTokenStream("Expected Expression.", fi)
#         condition = _expression_sa(loop_condition_tokens)
#         if not three_expressions:
#             fi = closing_parentheses.file_info
#             raise UnexpectedEndOfTokenStream("Expected Expression.", fi)
#         post_loop = _expression_sa(three_expressions)
#         if tokens[0].value == '{':
#             code = _code_block_sa(_get_nested_group(tokens, ('{','}'))[1])
#         else:
#             code = [_statement_sa(tokens)]
#         else_block = ElseBlock._sa(tokens)
#         if else_block is not None:
#             fi = stoken.file_info + else_block.file_info
#         else:
#             fi = stoken.file_info + code[-1].file_info
#         return ForBlock(pre_loop, condition, code, post_loop, else_block, fi)


# class WhileBlock:

#     _condition: Expression
#     _code: list[Statement]
#     _else: ElseBlock | None
#     _file_info: FileInfo

#     def __init__(
#         self,
#         condition: Expression,
#         code: list[Statement],
#         else_block: ElseBlock | None,
#         file_info: FileInfo,
#     ):
#         self._condition = condition
#         self._code = code[:]
#         self._else = else_block
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} While Loop\n"
#         if self._code or self._else is not None:
#             s += f"{pre_cont}├─ Condition\n"
#             cond_pre = f"{pre_cont}│ "
#         else:
#             s += f"{pre_cont}└─ Condition\n"
#             cond_pre = f"{pre_cont}  "
#         s += self._condition.tree_str(cond_pre + "└─", cond_pre + "  ")
#         if self._code:
#             if self._else is not None:
#                 s += f"{pre_cont}├─ Code\n"
#                 code_pre = f"{pre_cont}│ "
#             else:
#                 s += f"{pre_cont}└─ Code\n"
#                 code_pre = f"{pre_cont}  "
#             for code in self._code[:-1]:
#                 s += code.tree_str(code_pre + "├─", code_pre + "│ ")
#             s += self._code[-1].tree_str(code_pre + "└─", code_pre + "  ")
#         if self._else is not None:
#             s += self._else.tree_str(pre_cont + "└─", pre_cont + "  ")
#         return s

#     @staticmethod
#     def _sa(tokens: list[Token], token: Token) -> "WhileBlock":
#         _, condition_tokens, closing_parentheses = _get_nested_group(tokens)
#         if not condition_tokens:
#             fi = closing_parentheses.file_info
#             raise UnexpectedEndOfTokenStream("Expected Expression.", fi)
#         condition = _expression_sa(condition_tokens)
#         if tokens[0].value == '{':
#             code_tokens = _get_nested_group(tokens, ('{','}'))[1]
#             code = _code_block_sa(code_tokens)
#         else:
#             code = [_statement_sa(tokens)]
#         else_block = ElseBlock._sa(tokens)
#         if else_block is not None:
#             fi = token.file_info + else_block.file_info
#         else:
#             fi = token.file_info + code[-1].file_info
#         return WhileBlock(condition, code, else_block, fi)


# class DoBlock:

#     _first_code: list[Statement]
#     _condition: Expression
#     _second_code: list[Statement] | None
#     _else: ElseBlock | None
#     _file_info: FileInfo

#     def __init__(
#         self,
#         first_code: list[Statement],
#         condition: Expression,
#         second_code: list[Statement] | None,
#         else_block: ElseBlock | None,
#         file_info: FileInfo,
#     ):
#         self._first_code = first_code[:]
#         self._condition = condition
#         if second_code:
#             self._second_code = second_code[:]
#         else:
#             self._second_code = None
#         self._else = else_block
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Do Loop\n"
#         if self._first_code:
#             s += f"{pre_cont}├─ First Code\n"
#             code_pre = f"{pre_cont}│ "
#             for code in self._first_code[:-1]:
#                 s += code.tree_str(code_pre + "├─", code_pre + "│ ")
#             s += self._first_code[-1].tree_str(
#                 code_pre + "└─", code_pre + "  ")
#         if self._second_code or self._else is not None:
#             s += f"{pre_cont}├─ Condition\n"
#             cond_pre = f"{pre_cont}│ "
#         else:
#             s += f"{pre_cont}└─ Condition\n"
#             cond_pre = f"{pre_cont}  "
#         s += self._condition.tree_str(cond_pre + "└─", cond_pre + "  ")
#         if self._second_code:
#             if self._else is not None:
#                 s += f"{pre_cont}├─ Second Code\n"
#                 code_pre = f"{pre_cont}│ "
#             else:
#                 s += f"{pre_cont}└─ Second Code\n"
#                 code_pre = f"{pre_cont}  "
#             for code in self._second_code[:-1]:
#                 s += code.tree_str(code_pre + "├─", code_pre + "│ ")
#             s += self._second_code[-1].tree_str(
#                 code_pre + "└─", code_pre + "  ")
#         if self._else is not None:
#             s += self._else.tree_str(pre_cont + "└─", pre_cont + "  ")
#         return s

#     @staticmethod
#     def _sa(tokens: list[Token], stoken: Token) -> "DoBlock":
#         if tokens[0].value == '{':
#             code1_tokens = _get_nested_group(tokens, ('{','}'))[1]
#             code1 = _code_block_sa(code1_tokens)
#         else:
#             code1 = [_statement_sa(tokens)]
#         token = tokens.pop(0)
#         _assert_token(ExpectedKeyword, token, 'while')
#         _, condition_tokens, closing_parentheses = _get_nested_group(tokens)
#         if not condition_tokens:
#             fi = closing_parentheses.file_info
#             raise UnexpectedEndOfTokenStream("Expected Expression.", fi)
#         last_token = condition_tokens[-1]
#         condition = _expression_sa(condition_tokens)
#         if tokens[0].value == '{':
#             code2_tokens = _get_nested_group(tokens, ('{','}'))[1]
#             last_token = code2_tokens[-1]
#             code2 = _code_block_sa(code2_tokens)
#         elif tokens[0].value != 'else':
#             last_token = tokens[0]
#             code2 = [_statement_sa(tokens)]
#             if isinstance(code2[0], NoOperation):
#                 code2 = None
#         else:
#             code2 = None
#         else_block = ElseBlock._sa(tokens)
#         if else_block is not None:
#             fi = stoken.file_info + else_block.file_info
#         else:
#             fi = stoken.file_info + last_token.file_info
#         return DoBlock(code1, condition, code2, else_block, fi)


# class IfBlock:

#     _condition: Expression
#     _code: list[Statement]
#     _else: ElseBlock | None
#     _file_info: FileInfo

#     def __init__(
#         self,
#         condition: Expression,
#         code: list[Statement],
#         else_block: ElseBlock | None,
#         file_info: FileInfo,
#     ):
#         self._condition = condition
#         self._code = code[:]
#         self._else = else_block
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} If Statement\n"
#         if self._code or self._else is not None:
#             s += f"{pre_cont}├─ Condition\n"
#             cond_pre = f"{pre_cont}│ "
#         else:
#             s += f"{pre_cont}└─ Condition\n"
#             cond_pre = f"{pre_cont}  "
#         s += self._condition.tree_str(cond_pre + "└─", cond_pre + "  ")
#         if self._code:
#             if self._else is not None:
#                 s += f"{pre_cont}├─ Code\n"
#                 code_pre = f"{pre_cont}│ "
#             else:
#                 s += f"{pre_cont}└─ Code\n"
#                 code_pre = f"{pre_cont}  "
#             for code in self._code[:-1]:
#                 s += code.tree_str(code_pre + "├─", code_pre + "│ ")
#             s += self._code[-1].tree_str(code_pre + "└─", code_pre + "  ")
#         if self._else is not None:
#             s += self._else.tree_str(pre_cont + "└─", pre_cont + "  ")
#         return s

#     @staticmethod
#     def _sa(tokens: list[Token], token: Token) -> "IfBlock":
#         _, condition_tokens, closing_parentheses = _get_nested_group(tokens)
#         if not condition_tokens:
#             fi = closing_parentheses.file_info
#             raise UnexpectedEndOfTokenStream("Expected Expression.", fi)
#         condition = _expression_sa(condition_tokens)
#         if tokens[0].value == '{':
#             code = _code_block_sa(_get_nested_group(tokens, ('{','}'))[1])
#         else:
#             code = [_statement_sa(tokens)]
#         else_block = ElseBlock._sa(tokens)
#         if else_block is not None:
#             fi = token.file_info + else_block.file_info
#         else:
#             fi = token.file_info + code[-1].file_info
#         return IfBlock(condition, code, else_block, fi)


# class FunctionParameter:

#     _identifier: Identifier
#     _type: DataType
#     _pointer: bool
#     _default: Literal | None
#     _file_info: FileInfo

#     def __init__(
#         self,
#         identifier: Identifier,
#         type: DataType,
#         pointer: bool,
#         default: Literal | None,
#         file_info: FileInfo,
#     ):
#         self._identifier = identifier
#         self._type = type
#         self._pointer = pointer
#         self._default = default
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     @property
#     def identifier(self) -> Identifier: return self._identifier

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Function Parameter: {self._identifier}\n"
#         s += pre_cont
#         s += '├─ Type: ' if self._default else '└─ Type: '
#         if self._pointer: s+= "@"
#         s += f"{self._type}\n"
#         if self._default:
#             s += f"{pre_cont}└─ Default Value: {self._default}\n"
#         return s


# class FunctionBlock:

#     _identifier: Identifier
#     _params: list[FunctionParameter]
#     _return_type_pointer: bool
#     _return_type: DataType | None
#     _code: list[Statement]
#     _file_info: FileInfo

#     def __init__(
#         self,
#         identifier: Identifier,
#         params: list[FunctionParameter],
#         return_type_pointer: bool,
#         return_type: DataType | None,
#         code: list[Statement],
#         file_info: FileInfo,
#     ):
#         self._identifier = identifier
#         self._params = params[:]
#         self._return_type_pointer = return_type_pointer
#         self._return_type = return_type
#         self._code = code[:]
#         self._file_info = file_info

#     @property
#     def identifier(self) -> Identifier: return self._identifier

#     @property
#     def params(self) -> list[FunctionParameter]: return self._params[:]

#     @property
#     def return_type_pointer(self) -> bool: return self._return_type_pointer

#     @property
#     def return_type(self) -> DataType | None: return self._return_type

#     @property
#     def code(self) -> list[Statement]: return self._code[:]

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Function: {self._identifier}\n"
#         if self._params:
#             if self._code or self._return_type is not None:
#                 s += f"{pre_cont}├─ Parameters\n"
#                 params_pre = f"{pre_cont}│ "
#             else:
#                 s += f"{pre_cont}└─ Parameters\n"
#                 params_pre = f"{pre_cont}  "
#             for param in self._params[:-1]:
#                 s += param.tree_str(params_pre + "├─", params_pre + "│ ")
#             s += self._params[-1].tree_str(params_pre + "└─", params_pre + "  ")
#         if self._return_type is not None:
#             if self._code:
#                 s += f"{pre_cont}├─ Return Type: "
#             else:
#                 s += f"{pre_cont}└─ Return Type: "
#             if self._return_type_pointer: s+= "@"
#             s += f"{self._return_type}\n"
#         if self._code:
#             s += f"{pre_cont}└─ Code\n"
#             for code in self._code[:-1]:
#                 s += code.tree_str(pre_cont + "  ├─", pre_cont + "  │ ")
#             s += self._code[-1].tree_str(pre_cont + "  └─", pre_cont + "    ")
#         return s

#     @staticmethod
#     def _sa(tokens: list[Token], stoken: Token) -> "FunctionBlock":
#         identifier = tokens.pop(0)
#         _assert_token(ExpectedIdentifier, identifier)
#         token = tokens.pop(0)
#         _assert_token(ExpectedPunctuation, token, '(')
#         params: list[FunctionParameter] = []
#         while token.value != ')':
#             token = tokens.pop(0)
#             if isinstance(token, Punctuation):
#                 _assert_token(ExpectedPunctuation, token, ')')
#             elif isinstance(token, Identifier):
#                 member_id = Identifier(token.value, token.file_info)
#                 token = tokens.pop(0)
#                 _assert_token(ExpectedPunctuation, token, ':')
#                 pointer, data_type = _data_type_sa(tokens)
#                 token = tokens.pop(0)
#                 _assert_token(ExpectedPunctuation, token)
#                 if token.value not in [',', '=', ')']:
#                     raise UnexpectedPunctuation(token, [',', '=', ')'])
#                 elif token.value == '=':
#                     token = tokens.pop(0)
#                     _assert_token_literal(token)
#                     literal = _literal_map(token) # type: ignore
#                     token = tokens.pop(0)
#                     _assert_token(ExpectedPunctuation, token)
#                     if token.value not in [',', ')']:
#                         raise UnexpectedPunctuation(token, [',', ')'])
#                 else: literal = None
#                 if literal is not None:
#                     fi = member_id.file_info + literal.file_info
#                 else:
#                     fi = member_id.file_info + data_type.file_info
#                 params.append(FunctionParameter(
#                     member_id, data_type, pointer, literal, fi))
#             else:
#                 raise UnexpectedToken(
#                     token, ["Keyword", "Identifier", "Punctuation"])
#         token = tokens.pop(0)
#         _assert_token(ExpectedPunctuation, token, '->')
#         pointer, return_type = _data_type_sa(tokens)
#         code = _code_block_sa(_get_nested_group(tokens, ('{','}'))[1])
#         fi = token.file_info + code[-1].file_info
#         return FunctionBlock(
#             Identifier(identifier.value, identifier.file_info),
#             params,
#             pointer,
#             return_type,
#             code,
#             fi,
#         )


# class EnumMember:

#     _identifier: Identifier
#     _value: NumberLiteral | None
#     _file_info: FileInfo

#     def __init__(
#         self,
#         identifier: Identifier,
#         value: NumberLiteral | None,
#         file_info: FileInfo,
#     ):
#         self._identifier = identifier
#         self._value = value
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     @property
#     def identifier(self) -> Identifier: return self._identifier

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Enum Member: {self._identifier}\n"
#         if self._value is not None:
#             s += f"{pre_cont}└─ Value: {self._value}\n"
#         return s


# class EnumBlock:

#     _identifier: Identifier
#     _members: list[EnumMember]
#     _file_info: FileInfo

#     def __init__(
#         self,
#         identifier: Identifier,
#         members: list[EnumMember],
#         file_info: FileInfo,
#     ):
#         self._identifier = identifier
#         self._members = members[:]
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     @property
#     def identifier(self) -> Identifier: return self._identifier

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Enum: {self._identifier}\n"
#         if self._members:
#             for member in self._members[:-1]:
#                 s += member.tree_str(pre_cont + "├─", pre_cont + "│ ")
#             s += self._members[-1].tree_str(pre_cont + "└─", pre_cont + "  ")
#         return s

#     @staticmethod
#     def _sa(tokens: list[Token], stoken: Token) -> "EnumBlock":
#         identifier = tokens.pop(0)
#         _assert_token(ExpectedIdentifier, identifier)
#         token = tokens.pop(0)
#         _assert_token(ExpectedPunctuation, token, '{')
#         members: list[EnumMember] = []
#         while token.value != '}':
#             token = tokens.pop(0)
#             _assert_token(ExpectedIdentifier, token)
#             member_id = Identifier(token.value, token.file_info)
#             token = tokens.pop(0)
#             _assert_token(ExpectedPunctuation, token)
#             if token.value not in [',', '=', '}']:
#                 raise UnexpectedPunctuation(token, [',', '=', '}'])
#             elif token.value == '=':
#                 token = tokens.pop(0)
#                 _assert_token(ExpectedNumberLiteral, token)
#                 literal = NumberLiteral(token.value, token.file_info)
#                 token = tokens.pop(0)
#                 _assert_token(ExpectedPunctuation, token)
#                 if token.value not in [',', '}']:
#                     raise UnexpectedPunctuation(token, [',', '}'])
#             else: literal = None
#             if literal is not None:
#                 fi = member_id.file_info + literal.file_info
#             else:
#                 fi = member_id.file_info
#             members.append(EnumMember(member_id, literal, fi))
#         fi = stoken.file_info + token.file_info
#         return EnumBlock(
#             Identifier(identifier.value, identifier.file_info),
#             members,
#             fi,
#         )


# class StructureMember:

#     _identifier: Identifier
#     _type: DataType
#     _pointer: bool
#     _static: bool
#     _default: Literal | None
#     _file_info: FileInfo

#     def __init__(
#         self,
#         identifier: Identifier,
#         type: DataType,
#         pointer: bool,
#         static: bool,
#         default: Literal | None,
#         file_info: FileInfo,
#     ):
#         self._identifier = identifier
#         self._type = type
#         self._pointer = pointer
#         self._static = static
#         self._default = default
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     @property
#     def identifier(self) -> Identifier: return self._identifier

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Struct Member: {self._identifier}\n"
#         s += pre_cont
#         s += '├─ Type: ' if self._default else '└─ Type: '
#         if self._static: s+= "static "
#         if self._pointer: s+= "@"
#         s += f"{self._type}\n"
#         if self._default is not None:
#             s += f"{pre_cont}└─ Default Value: {self._default}\n"
#         return s


# class StructBlock:

#     _identifier: Identifier
#     _members: list[StructureMember]
#     _file_info: FileInfo

#     def __init__(
#         self,
#         identifier: Identifier,
#         members: list[StructureMember],
#         file_info: FileInfo,
#     ):
#         self._identifier = identifier
#         self._members = members[:]
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     @property
#     def identifier(self) -> Identifier: return self._identifier

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         s: str = f"{pre} Struct: {self._identifier}\n"
#         if self._members:
#             for member in self._members[:-1]:
#                 s += member.tree_str(pre_cont + "├─", pre_cont + "│ ")
#             s += self._members[-1].tree_str(pre_cont + "└─", pre_cont + "│ ")
#         return s

#     @staticmethod
#     def _sa(tokens: list[Token], stoken: Token) -> "StructBlock":
#         identifier = tokens.pop(0)
#         _assert_token(ExpectedIdentifier, identifier)
#         token = tokens.pop(0)
#         _assert_token(ExpectedPunctuation, token, '{')
#         members: list[StructureMember] = []
#         while token.value != '}':
#             token = tokens.pop(0)
#             if isinstance(token, Keyword):
#                 _assert_token(ExpectedKeyword, token, 'static')
#                 static_token = token
#                 token = tokens.pop(0)
#                 static = True
#             else:
#                 static = False
#             if isinstance(token, Identifier):
#                 member_id = Identifier(token.value, token.file_info)
#                 token = tokens.pop(0)
#                 _assert_token(ExpectedPunctuation, token, ':')
#                 pointer, data_type = _data_type_sa(tokens)
#                 token = tokens.pop(0)
#                 _assert_token(ExpectedPunctuation, token)
#                 if token.value not in [',', '=', '}']:
#                     raise UnexpectedPunctuation(token, [',', '=', '}'])
#                 elif token.value == '=':
#                     token = tokens.pop(0)
#                     _assert_token_literal(token)
#                     literal = _literal_map(token) # type: ignore
#                     token = tokens.pop(0)
#                     _assert_token(ExpectedPunctuation, token)
#                     if token.value not in [',', '}']:
#                         raise UnexpectedPunctuation(token, [',', '}'])
#                 else: literal = None
#                 if literal is not None:
#                     if static:
#                         fi = static_token.file_info + literal.file_info
#                     else:
#                         fi = member_id.file_info + literal.file_info
#                 else:
#                     if static:
#                         fi = static_token.file_info + data_type.file_info
#                     else:
#                         fi = member_id.file_info + data_type.file_info
#                 members.append(StructureMember(
#                     member_id, data_type, pointer, static, literal, fi))
#             else:
#                 raise UnexpectedToken(token, ["Keyword", "Identifier"])
#         fi = stoken.file_info + token.file_info
#         return StructBlock(
#             Identifier(identifier.value, identifier.file_info),
#             members,
#             fi,
#         )


# class Directive:

#     _content: str
#     _file_info: FileInfo

#     def __init__(
#         self,
#         content: str,
#         file_info: FileInfo,
#     ):
#         self._content = content
#         self._file_info = file_info

#     @property
#     def file_info(self) -> FileInfo: return self._file_info

#     def tree_str(self, pre: str = "", pre_cont: str = "") -> str:
#         return f"{pre} Directive: {self._content}\n"


class File:

    _children: list[Screen | Graph | Animation | Statement]
    _file_info: FileInfo

    def __init__(
        self,
        children: list[Screen | Graph | Animation | Statement],
        file_info: FileInfo,
    ):
        self._children = children[:]
        self._file_info = file_info

    @property
    def children(self) -> list[
        Screen |
        Graph |
        Animation |
        Statement
    ]:
        return self._children[:]

    @property
    def file_info(self) -> FileInfo: return self._file_info

    def tree_str(self) -> str:
        s: str = " File\n"
        if self._children:
            for child in self._children[:-1]:
                s += child.tree_str("├─", "│ ")
            s += self._children[-1].tree_str("└─", "  ")
        return s

    @staticmethod
    def _sa(tokens: list[Token]) -> "File":
        children: list[Screen | Graph | Animation | Statement] = []
        file_fi: FileInfo = tokens[0].file_info + tokens[-1].file_info

        while tokens:
            token = tokens.pop(0)
            _assert_token_mult(token, (Keyword))
            if isinstance(token, Keyword):
                match token.value.lower():
                    case 'screen':
                        children.append(Screen._sa(tokens, token))
                    case 'graph':
                        children.append(Graph._sa(tokens, token))
                    case 'anim':
                        children.append(Anim._sa(tokens, token))
                    case _:
                        raise ExpectedKeyword(token, "struct', 'enum', or 'fn")
            else:
                raise UnexpectedToken(token, "directive' or 'keyword")

        return File(children, file_fi)


# def _assert_token(
#     exception: type[_ExpectedTokenBase],
#     token: Token,
#     value: str | None = None,
#     token_type: type[Token] | None = None,
# ):
#     if not isinstance(token, token_type or exception._token_type):
#         raise exception(token)
#     if value is not None and token.value != value:
#         raise exception(token, value)

# def _assert_token_mult(
#     token: Token,
#     token_type: tuple[type[Token], ...],
# ):
#     if not isinstance(token, token_type):
#         raise UnexpectedToken(
#             token,
#             [i.__name__ for i in token_type], # type: ignore
#             type(token).__name__,
#         )

# def _assert_token_literal(
#     token: Token,
# ):
#     token_types = (
#         Keyword,
#         NumberLiteral,
#         CharLiteral,
#         StringLiteral,
#     )
#     if not isinstance(token, token_types):
#         raise UnexpectedToken(
#             token,
#             [i.__name__ for i in token_types], # type: ignore
#             type(token).__name__,
#         )
#     if isinstance(token, Keyword):
#         if token.value not in BuiltInConstEnum:
#             raise UnexpectedKeyword(
#                 token, [i.value for i in BuiltInDataTypeEnum])

# def _literal_map(literal: (
#     Keyword |
#     NumberLiteral |
#     CharLiteral |
#     StringLiteral
# )) -> Literal:
#     if isinstance(literal, Keyword):
#         return BuiltInConst(BuiltInConstEnum(literal.value), literal.file_info)
#     elif isinstance(literal, NumberLiteral):
#         return NumberLiteral(literal.value, literal.file_info)
#     elif isinstance(literal, CharLiteral):
#         return CharLiteral(literal.value, literal.file_info)
#     elif isinstance(literal, StringLiteral):
#         return StringLiteral(literal.value, literal.file_info)

# def _assert_token_value(
#     token: Token,
# ):
#     token_types = (
#         Identifier,
#         Keyword,
#         NumberLiteral,
#         CharLiteral,
#         StringLiteral,
#     )
#     if not isinstance(token, token_types):
#         raise UnexpectedToken(
#             token,
#             [i.__name__ for i in token_types], # type: ignore
#             type(token).__name__,
#         )
#     if isinstance(token, Keyword):
#         if token.value not in BuiltInConstEnum:
#             raise UnexpectedKeyword(
#                 token,
#                 [i.value for i in BuiltInDataTypeEnum]
#             )

# def _value_map(literal: (
#     Identifier |
#     Keyword |
#     NumberLiteral |
#     CharLiteral |
#     StringLiteral
# )) -> Literal | Identifier:
#     if isinstance(literal, Identifier):
#         return Identifier(literal.value, literal.file_info)
#     elif isinstance(literal, Keyword):
#         return BuiltInConst(BuiltInConstEnum(literal.value), literal.file_info)
#     elif isinstance(literal, NumberLiteral):
#         return NumberLiteral(literal.value, literal.file_info)
#     elif isinstance(literal, CharLiteral):
#         return CharLiteral(literal.value, literal.file_info)
#     elif isinstance(literal, StringLiteral):
#         return StringLiteral(literal.value, literal.file_info)

# def _get_nested_group(
#     tokens: list[Token],
#     encloses: tuple[str, str] = ('(',')'),
# ) -> tuple[Token, list[Token], Token]:
#     first_token = tokens.pop(0)
#     _assert_token(ExpectedPunctuation, first_token, encloses[0])
#     nested = 1
#     expr_len = -1
#     for i in range(len(tokens)):
#         if tokens[i].value == encloses[0]: nested += 1
#         elif tokens[i].value == encloses[1]: nested -= 1
#         if nested == 0:
#             expr_len = i
#             break
#     else:
#         raise UnexpectedEndOfTokenStream(
#             f"Expected '{encloses[1]}' but found '{tokens[-1].value}'.",
#             tokens[-1].file_info,
#         )
#     expr_tokens = tokens[:expr_len]
#     last_token = tokens[expr_len]
#     del tokens[:expr_len+1]
#     return first_token, expr_tokens, last_token

# def _get_to_symbol(
#     tokens: list[Token],
#     symbols: str | Sequence[str] = ';',
# ) -> tuple[list[Token], Token]:
#     expr_len = -1
#     for i in range(len(tokens)):
#         if tokens[i].value in symbols:
#             expr_len = i
#             break
#     else:
#         raise UnexpectedEndOfTokenStream(
#             "Unexpected End of Token Stream.", tokens[-1].file_info)
#     expr_tokens = tokens[:expr_len]
#     last_token = tokens[expr_len]
#     del tokens[:expr_len+1]
#     return expr_tokens, last_token

# def _data_type_sa(tokens: list[Token]) -> tuple[bool, DataType]:
#     token = tokens.pop(0)
#     _assert_token_mult(token, (
#         Keyword,
#         Identifier,
#         Punctuation,
#     ))
#     if isinstance(token, Punctuation):
#         _assert_token(ExpectedPunctuation, token, '@')
#         pointer = True
#         token = tokens.pop(0)
#         _assert_token_mult(token, (Keyword, Identifier))
#     else:
#         pointer = False
#     if isinstance(token, Keyword):
#         if token.value not in BuiltInDataTypeEnum:
#             raise UnexpectedKeyword(
#                 token,
#                 [i.value for i in BuiltInDataTypeEnum],
#             )
#         return pointer, BuiltInDataType(
#             BuiltInDataTypeEnum(token.value), token.file_info)
#     else:
#         return pointer, Identifier(token.value, token.file_info)

# def _code_block_sa(tokens: list[Token]) -> list[Statement]:
#     code: list[Statement] = []
#     while tokens:
#         code.append(_statement_sa(tokens))
#     return code

# def _expression_sa(tokens: list[Token]) -> Expression:
#     if not tokens:
#         raise ExpressionError("Expected Expression.")
#     elif len(tokens) == 1:
#         token = tokens.pop(0)
#         _assert_token_value(token)
#         return _value_map(token) # type: ignore
#     elif tokens[0].value == '(' and tokens[-1].value == ')':
#         if not tokens[1:-1]:
#             fi = tokens[0].file_info + tokens[-1].file_info
#             raise UnexpectedEndOfTokenStream(
#                 "Expected expression between '(' and ')'.", fi)
#         return _expression_sa(tokens[1:-1])

#     max_operator: int = -1
#     max_operator_precedence: int = -1
#     nested = 0
#     for i, token in enumerate(tokens):
#         if token.value == '(': nested += 1
#         elif token.value == ')':
#             if nested == 0:
#                 raise UnexpectedPunctuation(token, "(' before ')", token.value)
#             nested -= 1
#         elif nested == 0 and isinstance(token, Punctuation):
#             for j, operator in reversed(list(enumerate(_Operator_Precedence))):
#                 if j <= max_operator_precedence:
#                     break
#                 elif operator.value == token.value:
#                     max_operator = i
#                     max_operator_precedence = j
#                     break

#     if max_operator == -1:
#         function_identifier = tokens.pop(0)
#         _assert_token(ExpectedIdentifier, function_identifier)
#         token = tokens.pop(0)
#         _assert_token(ExpectedPunctuation, token, '(')
#         function_args: list[FunctionArgument] = []
#         while tokens:
#             arg_tokens, last_token = _get_to_symbol(tokens, (',', ')'))
#             if arg_tokens:
#                 if len(arg_tokens) > 1 and arg_tokens[1].value == '=':
#                     _assert_token(ExpectedIdentifier, arg_tokens[0])
#                     arg_identifier = Identifier(
#                         arg_tokens[0].value,
#                         arg_tokens[0].file_info,
#                     )
#                     del arg_tokens[:2]
#                 else:
#                     arg_identifier = None
#                 if not arg_tokens:
#                     fi = last_token.file_info
#                     raise UnexpectedEndOfTokenStream("Expected Expression.", fi)
#                 expression = _expression_sa(arg_tokens)
#                 if arg_identifier is not None:
#                     fi = arg_identifier.file_info + expression.file_info
#                 else:
#                     fi = expression.file_info
#                 function_args.append(
#                     FunctionArgument(arg_identifier, expression, fi))
#         fi = function_identifier.file_info + last_token.file_info
#         return FunctionCall(
#             Identifier(
#                 function_identifier.value,
#                 function_identifier.file_info,
#             ),
#             function_args,
#             fi,
#         )

#     if (
#         tokens[max_operator].value in PostfixUnaryOperatorEnum and
#         max_operator == len(tokens) - 1
#     ):
#         operator = UnaryOperator(
#             PostfixUnaryOperatorEnum(tokens[max_operator].value),
#             tokens[max_operator].file_info,
#         )
#         if not tokens[:max_operator]:
#             fi = tokens[max_operator].file_info
#             raise UnexpectedEndOfTokenStream(
#                 f"Expected expression before '{tokens[max_operator].value}'.",
#                 fi,
#             )
#         expression = _expression_sa(tokens[:max_operator])
#         fi = expression.file_info + operator.file_info
#         return UnaryExpression(operator, expression, fi)
#     elif (
#         tokens[max_operator].value in PrefixUnaryOperatorEnum and
#         max_operator == 0
#     ):
#         operator = UnaryOperator(
#             PrefixUnaryOperatorEnum(tokens[max_operator].value),
#             tokens[max_operator].file_info,
#         )
#         if not tokens[max_operator + 1:]:
#             fi = tokens[max_operator].file_info
#             raise UnexpectedEndOfTokenStream(
#                 f"Expected expression after '{tokens[max_operator].value}'.",
#                 fi,
#             )
#         expression = _expression_sa(tokens[max_operator + 1:])
#         fi = operator.file_info + expression.file_info
#         return UnaryExpression(operator, expression, fi)
#     elif tokens[max_operator].value in BinaryOperatorEnum:
#         operator = BinaryOperator(
#             BinaryOperatorEnum(tokens[max_operator].value),
#             tokens[max_operator].file_info,
#         )
#         if not tokens[:max_operator]:
#             fi = tokens[max_operator].file_info
#             raise UnexpectedEndOfTokenStream(
#                 f"Expected expression before '{tokens[max_operator].value}'.",
#                 fi,
#             )
#         expression1 = _expression_sa(tokens[:max_operator])
#         if not tokens[max_operator + 1:]:
#             fi = tokens[max_operator].file_info
#             raise UnexpectedEndOfTokenStream(
#                 f"Expected expression after '{tokens[max_operator].value}'.",
#                 fi,
#             )
#         expression2 = _expression_sa(tokens[max_operator + 1:])
#         fi = expression1.file_info + expression2.file_info
#         return BinaryExpression(operator, expression1, expression2, fi)
#     elif tokens[max_operator].value in TernaryOperatorEnum:
#         if not tokens[:max_operator]:
#             fi = tokens[max_operator].file_info
#             raise UnexpectedEndOfTokenStream(
#                 f"Expected expression before '{tokens[max_operator].value}'.",
#                 fi,
#             )
#         condition = _expression_sa(tokens[:max_operator])
#         del tokens[:max_operator]
#         operator = TernaryOperator(
#             TernaryOperatorEnum.TernaryConditional, tokens[0].file_info)
#         first_op, true_tokens, second_op = _get_nested_group(tokens, ('?', ':'))
#         if not true_tokens:
#             fi = first_op.file_info + second_op.file_info
#             raise UnexpectedEndOfTokenStream(
#                 "Expected expression between "
#                 f"'{first_op.value}' and '{second_op.value}'.",
#                 fi,
#             )
#         true_expr = _expression_sa(true_tokens)
#         if not tokens:
#             fi = second_op.file_info
#             raise UnexpectedEndOfTokenStream(
#                 f"Expected expression after '{second_op.value}'.",
#                 fi,
#             )
#         false_expr = _expression_sa(tokens)
#         fi = condition.file_info + false_expr.file_info
#         return TernaryExpression(operator, condition, true_expr, false_expr, fi)
#     else: raise SyntaxError(
#             "Expression Error", tokens[max_operator].file_info)

# def _statement_sa(tokens: list[Token]) -> Statement:
#     token = tokens.pop(0)
#     if isinstance(token, Keyword):
#         match token.value:
#             case 'let' | 'static':
#                 return LetStatement._sa(tokens, token)
#             case 'break' | 'continue' as key:
#                 fi = token.file_info
#                 token = tokens.pop(0)
#                 _assert_token(ExpectedPunctuation, token, ';')
#                 return LoopStatements(
#                     LoopStatementsEnum(key), fi + token.file_info)
#             case 'if':
#                 return IfBlock._sa(tokens, token)
#             case 'do':
#                 return DoBlock._sa(tokens, token)
#             case 'while':
#                 return WhileBlock._sa(tokens, token)
#             case 'for':
#                 return ForBlock._sa(tokens, token)
#             case key if key not in BuiltInConstEnum:
#                 raise UnexpectedKeyword(token, [
#                     'static',
#                     'let',
#                     'break',
#                     'continue',
#                     'if',
#                     'do',
#                     'while',
#                     'for',
#                 ] + [i.value for i in BuiltInConstEnum])
#     elif token.value == ';':
#         return NoOperation(token.file_info)
#     expr_tokens: list[Token] = [token] + _get_to_symbol(tokens)[0]
#     return _expression_sa(expr_tokens)

def syntactical_analyzer(tokens: Sequence[Token]) -> File:
    return File._sa(list(tokens))

if __name__ == '__main__':
    with open("example.graph", encoding='utf-8') as file:
        code = file.read()
    tokens = lexer(code, "example.graph")
    with open("tokens.txt", 'w', encoding='utf-8') as file:
        file.write('\n'.join([str(t) for t in tokens]))
